from datetime import datetime, timezone, timedelta

year_format = ["%Y", "%y"]
month_year_format = ["%m/%y", "%m/%Y", "%M/%Y", "%b/%Y", "%B/%y"]
day_month_year_format = ["%d/%m/%Y"]

slash_formats = day_month_year_format + year_format + month_year_format
dash_formats = [fmt.replace("/", "-") for fmt in slash_formats]

allowed_fmts = slash_formats + dash_formats


def parse_date(date, api, start=True):
    assert isinstance(date, str), "You didn't give a string as a date."
    for fmt in allowed_fmts:
        try:
            if start:
                date_object = datetime.strptime(date, fmt)
            else:
                # if an end date,
                # "2011"    -> "31-12-2011"
                # "01-2011" -> "31-01-2011"
                # "02-01-2011" -> "02-01-2011"
                date = datetime.strptime(date, fmt)
                date_object = date.replace(
                    month=12 if fmt in year_format else date.month,
                    day=last_day_of_month(date) if fmt.find("%d") == -1 else date.day,
                )
            return api_date_format_of(date_object, api)  # handle different api date formats
        except ValueError:
            continue
    raise ValueError


def parse_dates(start, end, api):
    """
    Auxiliary function to convert different date formats
    to %d/%m/%Y, which is required by the SGS API.
    """
    start = parse_date(start, api, start=True) if start else ""
    end = parse_date(end, api, start=False) if end else ""
    return start, end


def api_date_format_of(date, api):
    if api == "ipeadata":
        utc_offset = datetime.now(timezone.utc).astimezone().isoformat()[-6:]
        return date.strftime("%Y-%m-%dT00:00:00") + utc_offset  # 01-12-2010T00:00:00-03:00
    if api == "bcb":
        return date.strftime("%d/%m/%Y")  # 01/12/2010


def last_day_of_month(date):
    """
    Auxiliary function that returns the last day of the month,
    if no day is specified.

    See: https://stackoverflow.com/questions/42950/get-last-day-of-the-month
    """
    next_month = date.replace(day=28) + timedelta(days=4)
    return (next_month - timedelta(days=next_month.day)).day



import pandas as pd
from datetime import datetime


def parse_response(response, code, name, source):
    if source == "bcb":
        return parse_bcb_json(response, code, name)
    elif source == "ipea":
        return parse_ipea_json(response, code, name)


def parse_bcb_json(response, code, name):
    json = response.json()
    assert json, print(f"Request for {code} returned nothing.")
    return json_to_dataframe(json, code, name, "data", "valor", "%d/%m/%Y")


def parse_ipea_json(response, code, name):
    json = response.json()["value"]
    assert json, print(f"Request for {code} returned nothing.")
    return json_to_dataframe(
        json, code, name, "VALDATA", "VALVALOR", "%Y-%m-%dT%H:%M:%S%z"
    )


def json_to_dataframe(json, code, name, date_key, value_key, date_fmt):
    D = {"date": [], "value": []}
    date_value_generator = ((item[date_key], item[value_key]) for item in json)
    for date, value in date_value_generator:
        D["date"].append(datetime.strptime(date, date_fmt).strftime("%d/%m/%Y"))
        D["value"].append(value)
    dataframe = pd.DataFrame(D)
    dataframe["date"] = pd.to_datetime(dataframe["date"], format="%d/%m/%Y")
    dataframe["value"] = dataframe["value"].astype("float64")
    dataframe = dataframe.set_index("date").rename(
        columns={"value": name if name else code}
    )
    return dataframe



import os
import sys
import unittest
import pandas

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

from seriesbr import bcb


class BCBtest(unittest.TestCase):

    df = bcb.get_series(
        {
            "Spread": 20786,
            "Selic": 4189,
            "PIB_mensal": 4380,
            "Meta": 13521,
            "Dívida/PIB": 11405,
        }
    )

    def test_if_get_series_returns_data_frame(self):
        self.assertIsInstance(self.df, pandas.DataFrame)

    def test_if_search_empty_string_returns_data_frame(self):
        self.assertIsInstance(bcb.search(""), pandas.DataFrame)

    def test_if_search_returns_data_frame(self):
        self.assertIsInstance(bcb.search("Spread"), pandas.DataFrame)

    def test_if_search_more_than_one__params_returns_data_frame(self):
        self.assertIsInstance(bcb.search("Spread", "Pontos percentuais", "Mensal"), pandas.DataFrame)

    def test_search_with_argument_rows(self):
        self.assertEqual(len(bcb.search("Spread")), 10)

    def test_column_names(self):
        self.assertListEqual(self.df.columns.to_list(), ["Spread", "Selic", "PIB_mensal", "Meta", "Dívida/PIB"])

    def test_if_get_metadata_works(self):
        self.assertIsInstance(bcb.get_metadata(20786), dict)

    def test_if_last_n_works(self):
        self.assertEqual(len(bcb.get_series(20786, 4189, 4380, last_n=30)), 30)


if __name__ == "__main__":
    unittest.main()



from pprint import pprint


def ipea_make_select_query(fields):
    # to get the string
    # SERCODIGO,PERNOME,UNINOME,SERNOME,ANOTHERFILTER,ANOTHERFILTER
    # where ANOTHER must be something not alreay selected in default
    defaults = ["SERCODIGO", "SERNOME", "PERNOME", "UNINOME"]
    selected = defaults + [field for field in fields if field not in defaults]
    return f"?$select={','.join(selected)}"


def ipea_make_filter_query(name, fields):
    # to get the string "&$filter=contains(SERNOME,'name')
    # and contains(ANOTHER,'value') and contains(ANOTHER,'value')"
    filter_query = f"&$filter=contains(SERNOME,'{name}')"
    if any([field not in ipea_metadata_list for field in fields]):
        print_suggestions()
        raise ValueError("Not a valid field")
    if fields:
        filter_arguments = " and contains" + " and contains".join(
            f"({metadata},'{value}')" for metadata, value in fields.items()
        )
        return f"{filter_query}{filter_arguments}"
    return f"{filter_query}"


def print_suggestions():
    newline = "\n"
    print("These are not valid fields. Some hints:")
    pprint(ipea_metadata_list)


ipea_metadata_list = {
    "BASNOME": "Nome da base de dados da série.",
    "FNTNOME": "Nome completo da fonte da série, em português.",
    "FNTSIGLA": "Sigla ou nome abreviado da fonte da série, em português.",
    "FNTURL": "URL para o site da fonte da série.",
    "MULNOME": "Nome do fator multiplicador dos valores da série.",
    "PERNOME": "Nome da periodicidade, em português.",
    "SERATUALIZACAO": "Data da última carga de dados na série.",
    "SERCODIGO": "Código único de identificação da série.",
    "SERCOMENTARIO": "Comentários relativos a série, em português.",
    "SERNOME": "Nome da série, em português.",
    "UNINOME": "Nome da unidade dos valores da série.",
    "SERSTATUS": "Indica se uma série macroeconômica ainda é atualizada. Valores: ‘A’ (Ativa) para séries atualizadas ou ‘I’ (Inativa) para séries que não são atualizadas. As séries regionais ou sociais não possuem este metadado.",
    "TEMCODIGO": "Código de identificação do tema ao qual a série está associada.",
    "PAICODIGO": "Código de identificação país ou região (como América Latina, Zona do Euro, etc.) ao qual a série está associada. Deve ser levado em consideração apenas nas séries macroeconômicas (BASNOME = “Macroeconômico”); atualmente todas as séries regionais e sócias se referem ao Brasil, mesmo esta coluna PAICODIGO é nula ou vazia.",
    "SERNUMERICA": "Série numérica (1), série alfanumérica (0)",
}



from pandas import concat
from .helpers.dates import parse_dates
from .helpers.types import expect_type
from .helpers.response import parse_response
from .helpers.request import custom_get
from .helpers.search_results import return_search_results_bcb


def get_serie(code, name=None, start=None, end=None, last_n=None):
    """
    Returns a time series from Time Series Management System (SGS)
    of Brazilian Central Bank (BCB).

    Parameters
    ----------
    code (int or str): The code of the time series.

    name (str): The name of the series.

    start (str): Initial date.

    end (str): End date.

    last_n (int): Ignore other arguments and get the last
    n observations.

    Returns
    -------
    pandas.DataFrame
    """
    assert isinstance(code, str) or isinstance(code, int), "Not a valid code format."
    baseurl = f"http://api.bcb.gov.br/dados/serie/bcdata.sgs.{code}/dados"
    if last_n:
        url = f"{baseurl}/ultimos/{last_n - 1}?formato=json"
    else:
        start, end = parse_dates(start, end, api="bcb")
        dates = f"&dataInicial={start}&dataFinal={end}"
        url = f"{baseurl}?format=json{dates}"
    return parse_response(custom_get(url), code, name, source="bcb")


def get_series(*codes, start=None, end=None, last_n=None, **kwargs):
    """
    Get multiple series all at once in a single data frame.

    Parameters
    ----------
    codes (dict, str, int): dictionary like {"name1": cod1, "name2": cod2}
    or a bunch of code numbers like cod1, cod2.

    start (str): Initial date.

    end (str): End date.

    last_n (int): Get last_n observations.

    join (str): "outer" givess all observations and
    "inner" gives the intersection of them.

    **kwargs: passed to pandas.concat.

    Returns
    -------
    A DataFrame with the series.
    """
    assert codes, "You must pass at least one code."
    codes, names = expect_type(*codes)
    return concat(
        (get_serie(code, name, start, end, last_n) for code, name in zip(codes, names)),
        axis="columns",
        **kwargs
    )


def search(name="", *filters, rows=10, skip=1):
    """
    Search for a name in the SGS database.

    Parameters
    ----------
    name (str): string to search.

    skip (int or str): how many results to show.

    start (int or str): start showing from this result.

    Returns
    -------
    A DataFrame with the results.
    """
    # The api also takes parameters (filter queries)
    # but very often it returned nothing.
    # It turned out that, when I tried withou specifying
    # the queried parameters, I got better results in general
    # so I've sticked with this.
    # It doens't accept any keyword arguments, just positional arguments
    # that are then joined and passed to the &fq= API parameter.
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"q={name}&rows={rows}&start={skip}&sort=score desc"
    filter_params = "&fq=" + "+".join(f"{value}" for value in filters if filters)
    url = f"{baseurl}{params}{filter_params}"
    response = custom_get(url)
    results = return_search_results_bcb(response)
    return results


def get_metadata(code):
    """
    Get metadata of a series given its code.

    Parameters
    ----------
    code (str): code to search.

    Returns
    -------
    A dictionary with the results.
    """
    baseurl = "https://dadosabertos.bcb.gov.br/api/3/action/package_search?"
    params = f"fq=codigo_sgs:{code}"
    url = f"{baseurl}{params}"
    results = custom_get(url).json()["result"]["results"][0]
    return results



import re
from . import bcb, ipea, helpers
from pandas import concat


def get_series(*codes, start=None, end=None, **kwargs):
    codes, names = helpers.types.expect_type(*codes)
    series = []
    for code, name in zip(codes, names):
        if re.search(r"^\d+$", str(code)):
            series.append(bcb.get_serie(code, name, start, end))
        else:
            series.append(ipea.get_serie(code, name, start, end))
    return concat(series, axis="columns", **kwargs)



import os
import sys
import unittest
import pandas

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

from seriesbr import ipea


class IPEAtest(unittest.TestCase):

    series = {
        "IGP": "PAN12_IGPDIG12",
        "INAD": "BM12_CRLIN12",
        "IBC": "SGS12_IBCBRDESSAZ12",
    }

    df = ipea.get_series(series)

    def test_if_get_series_returns_data_frame(self):
        self.assertIsInstance(self.df, pandas.DataFrame)

    def test_if_search_empty_string_returns_data_frame(self):
        self.assertIsInstance(ipea.search(""), pandas.DataFrame)

    def test_column_names(self):
        self.assertListEqual(self.df.columns.to_list(), ["IGP", "INAD", "IBC"])

    def test_search_with_multiple_filters(self):
        result = ipea.search(BASNOME="Macroeconômico", PERNOME="Mensal", UNINOME="(p.p.)")
        self.assertGreater(len(result), 0)

    def test_if_metadata_returns_dict(self):
        self.assertIsInstance(ipea.get_metadata("PAN12_IGPDIG12"), dict)


if __name__ == "__main__":
    unittest.main()



from pandas import DataFrame


def return_search_results_bcb(response):
    json = response.json()
    search_results = json["result"]["results"]
    assert search_results, "Nothing was found."
    metadata_list = ["codigo_sgs", "title", "periodicidade", "unidade_medida"]
    D = {metadata: [] for metadata in metadata_list}
    for result in search_results:
        for metadata in metadata_list:
            D[metadata].append(result[metadata])
    df = DataFrame(D)
    return df


def return_search_results_ipea(response):
    search_results = response.json()["value"]
    assert search_results, "Nothing was found."
    D = {metadata: [] for metadata in search_results[0].keys()}
    for result in search_results:
        for metadata in result.keys():
            D[metadata].append(result[metadata])
    df = DataFrame(D)
    return df



import setuptools
import distutils

long_description = """
SeriesBR - A Python package to get brazilian economic time series

It has functions to get time series data from the Instituto de Pesquisa Econômica Aplicada (IPEA) and Brazilian Central Bank (BCB) databases into a pandas.DataFrame.

There are also functions to make queries and to get metadata information.

You can discover more about it in its GitHub [repository](https://github.com/phelipetls/seriesbr).
"""

setuptools.setup(
    name="seriesbr",
    version="0.0.3",
    author="Phelipe Teles",
    author_email="phelipe_teles@hotmail.com",
    description="Get requests for brazilian economic time series databases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phelipetls/brseries",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from pandas import concat, DataFrame
from .helpers.types import expect_type
from .helpers.request import custom_get
from .helpers.response import parse_response
from .helpers.search_results import return_search_results_ipea
from .helpers.dates import parse_dates
from .helpers.metadata import (
    ipea_make_select_query,
    ipea_make_filter_query,
)


def get_serie(code, name=None, start=None, end=None):
    """
    Returns a time series from IPEADATA database.

    Parameters
    ----------
    code (int or str): The code of the time series.

    name (str): The name of the series.

    start (str): Initial date.

    end (str): End date.

    Returns
    -------
    pandas.DataFrame
    """
    assert isinstance(code, str), "Not a valid code format."
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"ValoresSerie(SERCODIGO='{code}')"
    select = "?$select=VALDATA,VALVALOR"
    start, end = parse_dates(start, end, api="ipeadata")
    dates = date_filter(start, end)
    url = f"{baseurl}{resource_path}{select}{dates}"
    serie = parse_response(custom_get(url), code, name, source="ipea")
    return serie


def date_filter(start, end):
    """
    Auxiliary function to return the right query
    to filter dates.
    """
    if start and end:
        data = f"&$filter=VALDATA ge {start} and VALDATA le {end}"
    elif start:
        data = f"&$filter=VALDATA ge {start}"
    elif end:
        data = f"&$filter=VALDATA le {end}"
    else:
        data = ""
    return data


def get_series(*codes, start=None, end=None, **kwargs):
    """
    Get multiple series all at once in a single data frame.

    Parameters
    ----------
    codes (dict, str, int): dictionary like {"name1": cod1, "name2": cod2}
    or a bunch of code numbers like cod1, cod2.

    start (str): Initial date.

    end (str): End date.

    **kwargs: passed to pandas.concat.

    Returns
    -------
    pandas.DataFrame with the requested series.
    """
    assert codes, "You must pass at least one code to be searched."
    codes, names = expect_type(*codes)
    return concat(
        [get_serie(code, name, start, end) for code, name in zip(codes, names)],
        axis="columns",
        sort=True,
        **kwargs,
    )


def search(SERNOME="", **fields):
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = "Metadados"
    select_query = ipea_make_select_query(fields)
    filter_query = ipea_make_filter_query(SERNOME, fields)
    url = f"{baseurl}{resource_path}{select_query}{filter_query}"
    response = custom_get(url)
    results = return_search_results_ipea(response)
    return results


def get_metadata(cod):
    """
    Returns metadata of a series specified by cod.
    """
    baseurl = "http://ipeadata2-homologa.ipea.gov.br/api/v1/"
    resource_path = f"Metadados('{cod}')"
    url = f"{baseurl}{resource_path}"
    results = custom_get(url).json()["value"][0]
    return results


def list_themes():
    return DataFrame(list_themes("Temas", "TEMCODIGO", "TEMNOME"))


def list_countries():
    return DataFrame(list_metadata("Paises", "PAICODIGO", "PAINOME"))


def list_metadata(resource_path, code_key, value_key):
    baseurl = "http://www.ipeadata.gov.br/api/odata4/"
    url = f"{baseurl}{resource_path}"
    response = custom_get(url).json()["value"]
    return [{code_key: item[code_key], value_key: item[value_key]} for item in response]



import os
import sys
import unittest
import pandas

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))

from seriesbr import seriesbr


class SERIESBRtest(unittest.TestCase):

    df = seriesbr.get_series({"Spread": 20786, "Selic": 4189, "PIB_mensal": 4380})

    def test_get_series_returns_data_frame(self):
        self.assertIsInstance(self.df, pandas.DataFrame)

    def test_if_names_of_columns_are_right(self):
        self.assertListEqual(self.df.columns.to_list(), ["Spread", "Selic", "PIB_mensal"])


if __name__ == "__main__":
    unittest.main()
